@@ -2,6 +2,8 @@ import errno
 import logging
 import os
 from threading import Thread
+import time
+import sys
 
 import zmq
 from zmq.eventloop import ioloop
@@ -23,6 +25,8 @@ class Arbiter(object):
     - **watchers** -- a list of Watcher objects
     - **endpoint** -- the controller ZMQ endpoint
     - **pubsub_endpoint** -- the pubsub endpoint
+    - **stats_endpoint** -- the stats endpoint. If not provided,
+      the *circusd-stats* process will not be launched.
     - **check_delay** -- the delay between two controller points
       (default: 1 s)
     - **prereload_fn** -- callable that will be executed on each reload
@@ -36,7 +40,7 @@ class Arbiter(object):
     """
     def __init__(self, watchers, endpoint, pubsub_endpoint, check_delay=1.,
                  prereload_fn=None, context=None, loop=None,
-                 check_flapping=True):
+                 check_flapping=True, stats_endpoint=None):
         self.watchers = watchers
         self.endpoint = endpoint
         self.check_delay = check_delay
@@ -55,6 +59,17 @@ class Arbiter(object):
         self.busy = False
         self.check_flapping = check_flapping
 
+        # initializing circusd-stats as a watcher when configured
+        self.stats_endpoint = stats_endpoint
+        if self.stats_endpoint is not None:
+            cmd = "%s -c 'from circus import stats; stats.main()'" % \
+                        sys.executable
+            cmd += ' --endpoint %s' % self.endpoint
+            cmd += ' --pubsub %s' % self.pubsub_endpoint
+            cmd += ' --statspoint %s' % self.stats_endpoint
+            stats_watcher = Watcher('circusd-stats', cmd)
+            self.watchers.append(stats_watcher)
+
     @classmethod
     def load_from_config(cls, config_file):
         cfg = get_config(config_file)
@@ -69,10 +84,18 @@ class Arbiter(object):
         # creating arbiter
         arbiter = cls(watchers, cfg['endpoint'], cfg['pubsub_endpoint'],
                       check_delay=cfg.get('check_delay', 1.),
-                      prereload_fn=cfg.get('prereload_fn'))
+                      prereload_fn=cfg.get('prereload_fn'),
+                      stats_endpoint=cfg.get('stats_endpoint'))
 
         return arbiter
 
+    def iter_watchers(self):
+        watchers = [(watcher.priority, watcher) for watcher in self.watchers]
+        watchers.sort()
+        watchers.reverse()
+        for __, watcher in watchers:
+            yield watcher
+
     @debuglog
     def initialize(self):
         # set process title
@@ -89,7 +112,7 @@ class Arbiter(object):
                                      self.pubsub_endpoint, self.check_delay)
 
         # initialize watchers
-        for watcher in self.watchers:
+        for watcher in self.iter_watchers():
             self._watchers_names[watcher.name.lower()] = watcher
             watcher.initialize(self.evpub_socket)
 
@@ -115,7 +138,7 @@ class Arbiter(object):
 
         # initialize processes
         logger.debug('Initializing watchers')
-        for watcher in self.watchers:
+        for watcher in self.iter_watchers():
             watcher.start()
 
         logger.info('Arbiter now waiting for commands')
@@ -141,12 +164,40 @@ class Arbiter(object):
             self.stop_watchers(stop_alive=True)
         self.loop.stop()
 
+    def reap_processes(self):
+        # map watcher to pids
+        watchers_pids = {}
+        for watcher in self.iter_watchers():
+            if not watcher.stopped:
+                for pid, wid in watcher.pids.items():
+                    watchers_pids[pid] = (watcher, wid)
+
+        # detect dead children
+        while True:
+            try:
+                pid, status = os.waitpid(-1, os.WNOHANG)
+                if not pid:
+                    break
+
+                if pid in watchers_pids:
+                    watcher, wid = watchers_pids[pid]
+                    watcher.reap_process(wid, status)
+            except OSError as e:
+                if e.errno == errno.EAGAIN:
+                    time.sleep(0.001)
+                    continue
+                elif e.errno == errno.ECHILD:
+                    # process already reaped
+                    return
+                else:
+                    raise
+
     def manage_watchers(self):
         if not self.busy and self.alive:
             self.busy = True
             # manage and reap processes
-            for watcher in self.watchers:
-                watcher.reap_processes()
+            self.reap_processes()
+            for watcher in self.iter_watchers():
                 watcher.manage_processes()
 
             if self.check_flapping and not self.flapping.is_alive():
@@ -177,7 +228,7 @@ class Arbiter(object):
                 handler.release()
 
         # gracefully reload watchers
-        for watcher in self.watchers:
+        for watcher in self.iter_watchers():
             watcher.reload(graceful=graceful)
 
     def numprocesses(self):
@@ -196,13 +247,14 @@ class Arbiter(object):
         return dict([(watcher.name, watcher.status())
                       for watcher in self.watchers])
 
-    def add_watcher(self, name, cmd):
+    def add_watcher(self, name, cmd, **kw):
         """Adds a watcher.
 
         Options:
 
         - **name**: name of the watcher to add
         - **cmd**: command to run.
+        - all other options defined in the Watcher constructor.
         """
         if name in self._watchers_names:
             raise AlreadyExist("%r already exist" % name)
@@ -210,10 +262,11 @@ class Arbiter(object):
         if not name:
             return ValueError("command name shouldn't be empty")
 
-        watcher = Watcher(name, cmd, stopped=True)
+        watcher = Watcher(name, cmd, **kw)
         watcher.initialize(self.evpub_socket)
         self.watchers.append(watcher)
         self._watchers_names[watcher.name.lower()] = watcher
+        return watcher
 
     def rm_watcher(self, name):
         """Deletes a watcher.
@@ -232,7 +285,7 @@ class Arbiter(object):
         watcher.stop()
 
     def start_watchers(self):
-        for watcher in self.watchers:
+        for watcher in self.iter_watchers():
             watcher.start()
 
     def stop_watchers(self, stop_alive=False):
@@ -243,9 +296,13 @@ class Arbiter(object):
             logger.info('Arbiter exiting')
             self.alive = False
 
-        for watcher in self.watchers:
+        for watcher in self.iter_watchers():
             watcher.stop()
 
+    def restart(self):
+        self.stop_watchers()
+        self.start_watchers()
+
 
 class ThreadedArbiter(Arbiter, Thread):
 
