@@ -8,6 +8,7 @@ import sys
 import shlex
 
 from psutil.error import AccessDenied, NoSuchProcess
+from psutil import Process
 from circus import logger
 
 
@@ -89,7 +90,26 @@ def bytes2human(n):
     return "%sB" % n
 
 
-def get_info(process):
+# XXX weak dict ?
+_PROCS = {}
+
+
+def get_info(process=None, interval=0):
+    """Return information about a process. (can be an pid or a Process object)
+
+    If process is None, will return the information about the current process.
+    """
+    if process is None or isinstance(process, int):
+        if process is None:
+            pid = os.getpid()
+        else:
+            pid = process
+
+        if pid in _PROCS:
+            process = _PROCS[pid]
+        else:
+            _PROCS[pid] = process = Process(pid)
+
     info = {}
     try:
         mem_info = process.get_memory_info()
@@ -99,7 +119,7 @@ def get_info(process):
         info['mem_info1'] = info['mem_info2'] = "N/A"
 
     try:
-        info['cpu'] = process.get_cpu_percent(interval=0)
+        info['cpu'] = process.get_cpu_percent(interval=interval)
     except AccessDenied:
         info['cpu'] = "N/A"
 
@@ -143,6 +163,10 @@ def get_info(process):
 
     info['cmdline'] = cmdline
 
+    info['children'] = []
+    for child in process.get_children():
+        info['children'].append(get_info(child, interval=interval))
+
     return info
 
 
