@@ -1,10 +1,12 @@
 import errno
+import os
 import signal
 import time
 
+from psutil import STATUS_ZOMBIE, STATUS_DEAD
 from zmq.utils.jsonapi import jsonmod as json
 
-from circus.process import Process, DEAD_OR_ZOMBIE
+from circus.process import Process
 from circus import logger
 from circus import util
 from circus.stream import get_pipe_redirector
@@ -40,6 +42,9 @@ class Watcher(object):
     - **gid**: if given, is the group id or name the command should run
       with. The current gid is the default.
 
+    - **send_hup**: if True, a process reload will be done by sending
+      the SIGHUP signal. Defaults to False.
+
     - **env**: a mapping containing the environment variables the command
       will run with. Optional.
 
@@ -84,14 +89,28 @@ class Watcher(object):
       process. Can be *thread* or *gevent*. When set to *gevent* you need
       to have *gevent* and *gevent_zmq* installed. (default: thread)
 
+    - **flapping_attempts** -- number of times a process can restart before we
+      start to detect the flapping (default: 2)
+    - **flapping_window** -- the time window in seconds to test for flapping.
+      If the process restarts more than **times** times, we consider it a
+      flapping process. (default: 1)
+    - **retry_in**: time in seconds to wait until we try to start a process
+      that has been flapping. (default: 7)
+    - **max_retry**: the number of times we attempt to start a process, before
+      we abandon and stop the whole watcher. (default: 5)
+    - **priority**: integer that defines a priority for the watcher. When
+      the Arbiter do some operations on all watchers, it will sort them
+      with this field, from the bigger number to the smallest.
+      (default: 0)
     """
     def __init__(self, name, cmd, args=None, numprocesses=1, warmup_delay=0.,
                  working_dir=None, shell=False, uid=None,
                  gid=None, send_hup=False, env=None, stopped=True,
-                 times=2, within=1., retry_in=7., max_retry=5,
+                 flapping_attempts=2, flapping_window=1., retry_in=7.,
+                 max_retry=5,
                  graceful_timeout=30., prereload_fn=None,
                  rlimits=None, executable=None, stdout_stream=None,
-                 stderr_stream=None, stream_backend='thread'):
+                 stderr_stream=None, stream_backend='thread', priority=0):
         self.name = name
         self.res_name = name.lower().replace(" ", "_")
         self.numprocesses = int(numprocesses)
@@ -100,39 +119,32 @@ class Watcher(object):
         self.args = args
         self._process_counter = 0
         self.stopped = stopped
-        self.times = times
-        self.within = within
+        self.flapping_attempts = flapping_attempts
+        self.flapping_window = flapping_window
         self.retry_in = retry_in
         self.max_retry = max_retry
         self.graceful_timeout = 30
         self.prereload_fn = prereload_fn
         self.executable = None
         self.stream_backend = stream_backend
-
+        self.priority = priority
         self.stdout_stream = stdout_stream
-        if stdout_stream:
-            self.stdout_redirector = get_pipe_redirector(stdout_stream,
-                    backend=stream_backend)
-        else:
-            self.stdout_redirector = None
-
         self.stderr_stream = stderr_stream
-        if stderr_stream:
-            self.stderr_redirector = get_pipe_redirector(stderr_stream,
-                    backend=stream_backend)
-        else:
-            self.stderr_redirector = None
+        self.stdout_redirector = self.stderr_redirector = None
 
         self.optnames = ("numprocesses", "warmup_delay", "working_dir",
                          "uid", "gid", "send_hup", "shell", "env",
-                         "cmd", "times", "within", "retry_in", "args",
-                         "max_retry", "graceful_timeout", "executable")
+                         "cmd", "flapping_attempts", "flapping_window",
+                         "retry_in", "args",
+                         "max_retry", "graceful_timeout", "executable",
+                         "priority")
 
         if not working_dir:
             # working dir hasn't been set
             working_dir = util.get_working_dir()
 
         self.working_dir = working_dir
+        self.pids = {}
         self.processes = {}
         self.shell = shell
         self.uid = uid
@@ -142,6 +154,27 @@ class Watcher(object):
         self.send_hup = send_hup
         self.evpub_socket = None
 
+    def _create_redirectors(self):
+        if self.stdout_stream:
+            if (self.stdout_redirector is not None and
+                self.stdout_redirector.running):
+                self.stdout_redirector.kill()
+
+            self.stdout_redirector = get_pipe_redirector(self.stdout_stream,
+                    backend=self.stream_backend)
+        else:
+            self.stdout_redirector = None
+
+        if self.stderr_stream:
+            if (self.stderr_redirector is not None and
+                self.stderr_redirector.running):
+                self.stderr_redirector.kill()
+
+            self.stderr_redirector = get_pipe_redirector(self.stderr_stream,
+                    backend=self.stream_backend)
+        else:
+            self.stderr_redirector = None
+
     @classmethod
     def load_from_config(cls, config):
         return cls(name=config['name'],
@@ -156,8 +189,8 @@ class Watcher(object):
                    send_hup=config.get('send_hup', False),
                    env=config.get('env'),
                    stopped=config.get('stopped', True),
-                   times=config.get('times', 2),
-                   within=config.get('within', 1),
+                   flapping_attempts=config.get('flapping_attempts', 2),
+                   flapping_window=config.get('flapping_window', 1),
                    retry_in=config.get('retry_in', 7),
                    max_retry=config.get('max_retry', 5),
                    graceful_timeout=config.get('graceful_timeout', 30),
@@ -166,7 +199,8 @@ class Watcher(object):
                    executable=config.get('executable'),
                    stdout_stream=config.get('stdout_stream'),
                    stderr_stream=config.get('stderr_stream'),
-                   stream_backend=config.get('stream_backend', 'thread'))
+                   stream_backend=config.get('stream_backend', 'thread'),
+                   priority=int(config.get('priority', 0)))
 
     @util.debuglog
     def initialize(self, evpub_socket):
@@ -192,6 +226,31 @@ class Watcher(object):
         if not self.evpub_socket.closed:
             self.evpub_socket.send_multipart(multipart_msg)
 
+    @util.debuglog
+    def reap_process(self, wid, status):
+        process = self.processes.pop(wid)
+        self.pids.pop(process.pid)
+
+        # get retrn code
+        if os.WIFSIGNALED(status):
+            retcode = os.WTERMSIG(status)
+        # process exited using exit(2) system call; return the
+        # integer exit(2) system call has been called with
+        elif os.WIFEXITED(status):
+            retcode = os.WEXITSTATUS(status)
+        else:
+            # should never happen
+            raise RuntimeError("Unknown process exit status")
+
+        # if the process is dead or a zombie try to definitely stop it.
+        if retcode in (STATUS_ZOMBIE, STATUS_DEAD):
+            process.stop()
+
+        logger.debug("reap process %s", process.pid)
+        self.send_msg("reap", {"process_id": wid,
+                               "process_pid": process.pid,
+                               "time": time.time()})
+
     @util.debuglog
     def reap_processes(self):
         """Reap processes.
@@ -199,17 +258,27 @@ class Watcher(object):
         if self.stopped:
             return
 
-        for wid, process in self.processes.items():
-            if process.poll() is not None:
-                if process.status == DEAD_OR_ZOMBIE:
-                    process.stop()
+        while True:
+            try:
+                pid, status = os.waitpid(-1, os.WNOHANG)
+                if not pid:
+                    break
 
-                self.send_msg("reap", {"process_id": wid,
-                                       "process_pid": process.pid,
-                                       "time": time.time()})
+                if pid in self.pids:
+                    self.reap_process(self.pids[pid], status)
+
+                # watcher have been stopped, exit the loop
                 if self.stopped:
                     break
-                self.processes.pop(wid)
+            except OSError as e:
+                if e.errno == errno.EAGAIN:
+                    time.sleep(0.001)
+                    continue
+                elif e.errno == errno.ECHILD:
+                    # process already reaped
+                    return
+                else:
+                    raise
 
     @util.debuglog
     def manage_processes(self):
@@ -226,6 +295,9 @@ class Watcher(object):
         while len(processes) > self.numprocesses:
             wid = processes.pop(0)
             process = self.processes.pop(wid)
+            if process.pid in self.pids:
+                self.pids.pop(process.pid)
+
             self.kill_process(process)
 
     @util.debuglog
@@ -274,6 +346,7 @@ class Watcher(object):
                                                            process.stderr)
 
                 self.processes[self._process_counter] = process
+                self.pids[process.pid] = process.wid
                 logger.debug('running %s process [pid %d]', self.name,
                             process.pid)
             except OSError, e:
@@ -302,6 +375,7 @@ class Watcher(object):
             self.stderr_redirector.remove_redirection('stderr', process)
 
         self.send_msg("kill", {"process_id": process.wid,
+                               "process_pid": process.pid,
                                "time": time.time()})
         logger.debug("%s: kill process %s", self.name, process.pid)
         process.send_signal(sig)
@@ -314,6 +388,7 @@ class Watcher(object):
         for wid in self.processes.keys():
             try:
                 process = self.processes.pop(wid)
+                del self.pids[process.pid]
                 self.kill_process(process, sig)
             except OSError as e:
                 if e.errno != errno.ESRCH:
@@ -397,8 +472,10 @@ class Watcher(object):
             return
 
         self.stopped = False
+        self._create_redirectors()
         self.reap_processes()
         self.manage_processes()
+
         if self.stdout_redirector is not None:
             self.stdout_redirector.start()
 
@@ -489,11 +566,11 @@ class Watcher(object):
         elif key == "cmd":
             self.cmd = val
             action = 1
-        elif key == "times":
-            self.times = int(val)
+        elif key == "flapping_attempts":
+            self.flapping_attempts = int(val)
             action = -1
-        elif key == "within":
-            self.within = float(val)
+        elif key == "flapping_window":
+            self.flapping_window = float(val)
         elif key == "retry_in":
             self.retry_in = float(val)
         elif key == "max_retry":
